(*
Functional maps as balanced binary trees

D. Monniaux, 2003
Laboratoire d'Informatique de l'Ecole Normale Superieure
*)

Lemma bal_impossible1 : (tree : t)
  (n : nat) (constraint : (gt (height tree) n))
  (X : Set) (f : (ll : t) (lv : key) (ld : a) (lr : t) X) X.
Proof.
  Intro tree.
  Case tree.
  Intros n absurd.
  Absurd (gt (height Empty) n).
  Unfold height.
  Auto with arith.

  Assumption.

  Intros.
  Exact (f left k data right).
Defined.

Definition bal [l : t] [k : key] [data : a] [r : t] :=
  let hl = (height l) in
  let hr = (height r) in
  Cases (le_gt_dec hl (S (S hr))) of
  | (right absurd) =>
     (bal_impossible1 l  (S (S hr)) absurd t
     ([ll : t] [lv : key] [ld : a] [lr : t]
       Cases (le_gt_dec (height lr) (height ll)) of
       | (left _) => (create ll lv ld (create lr k data r))
       | (right absurd2) =>
         (bal_impossible1 lr (height ll) absurd2 t
	 ([lrl : t] [lrv : key] [lrd : a] [lrr : t]
           (create (create ll lv ld lrl) lrv lrd (create lrr k data r))))
       end))
  | (left _) =>
  Cases (le_gt_dec hr (S (S hl))) of
  | (right absurd) =>
     (bal_impossible1 r (S (S hl)) absurd t
     ([rl : t] [rv : key] [rd : a] [rr : t]
       Cases (le_gt_dec (height rl) (height rr)) of
       | (left _) => (create (create l k data rl) rv rd rr)
       | (right absurd2) =>
         (bal_impossible1 rl (height rr) absurd2 t
	 ([rll : t] [rlv : key] [rld : a] [rlr : t]
           (create (create l k data rll) rlv rld (create rlr rv rd rr))))
       end))
  | (left _) => (create l k data r)
  end end.

Lemma bal_wf_order:
  (left : t) (k : key) (data : a) (right : t)
  (left_wf : (wf_order left))
  (right_wf : (wf_order right))
  (left_bounded : (bounded_up k left))
  (right_bounded : (bounded_down k right))
  (wf_order (bal left k data right)).
Proof.
  Intros.
  Unfold bal.
  NewDestruct (le_gt_dec (height left) (S (S (height right)))).
  NewDestruct (le_gt_dec (height right) (S (S (height left)))).
  Apply create_wf_order; Assumption.

  NewDestruct right.
  Absurd (gt (height Empty) (S (S (height left)))).
  Unfold height.
  Auto with arith.

  Unfold height.
  Auto with arith.

  Unfold bal_impossible1.
  NewDestruct (le_gt_dec (height left0) (height right)).
  Apply create_wf_order.
  Apply create_wf_order.
  Assumption.

  Inversion right_wf.
  Assumption.

  Assumption.

  Inversion right_bounded.
  Assumption.

  Inversion right_wf.
  Assumption.

  Unfold create.
  Apply node_bounded_up.
  Inversion right_bounded.
  Apply wf_order_overbound with k:=k; Assumption.

  Inversion right_bounded.
  Apply compare_reverse1; Assumption.

  Inversion right_wf; Assumption.

  Inversion right_wf; Assumption.

  NewDestruct left0.
  Absurd (gt O (height right)).
  Auto with arith.

  Exact g0.

  Apply create_wf_order.
  Apply create_wf_order.
  Assumption.

  Inversion right_wf.
  Inversion left_wf0; Assumption.

  Assumption.

  Inversion right_bounded.
  Inversion left_bounded0; Assumption.

  Apply create_wf_order.
  Inversion right_wf.
  Inversion left_wf0; Assumption.

  Inversion right_wf; Assumption.

  Inversion right_wf.
  Inversion left_bounded0.
  Assumption.

  Inversion right_wf.
  Assumption.

  Unfold create.
  Constructor.
  Inversion right_bounded.
  Inversion left_bounded0.
  Apply wf_order_overbound with k:=k; Assumption.

  Inversion right_bounded.
  Inversion left_bounded0.
  Apply compare_reverse1.
  Assumption.

  Inversion right_wf.
  Inversion left_wf0.
  Assumption.

  Unfold create.
  Constructor.
  Inversion right_wf.
  Inversion left_wf0; Assumption.

  Inversion right_wf.
  Inversion left_bounded0.
  Apply compare_reverse2; Assumption.

  Apply wf_order_underbound with k:=k0.
  Inversion right_wf.
  Inversion left_bounded0.
  Assumption.

  Inversion right_wf.
  Assumption.

  NewDestruct left.
  Absurd (gt O (S (S (height right)))).
  Auto with arith.

  Exact g.

  Simpl.
  NewDestruct (le_gt_dec (height right0) (height left)).
  Apply create_wf_order.
  Inversion left_wf.
  Assumption.

  Apply create_wf_order.
  Inversion left_wf.
  Assumption.

  Assumption.

  Inversion left_bounded.
  Assumption.

  Assumption.

  Inversion left_wf.
  Assumption.

  Unfold create.
  Constructor.
  Inversion left_wf.
  Assumption.

  Inversion left_bounded.
  Apply compare_reverse2.
  Assumption.

  Apply wf_order_underbound with k:=k.
  Inversion left_bounded.
  Assumption.

  Assumption.

  NewDestruct right0.
  Absurd (gt O (height left)).
  Auto with arith.

  Assumption.

  Simpl.
  Apply create_wf_order.
  Apply create_wf_order.
  Inversion left_wf; Assumption.

  Inversion left_wf.
  Inversion right_wf0; Assumption.

  Inversion left_wf.
  Assumption.

  Inversion left_wf.
  Inversion right_bounded0.
  Assumption.

  Apply create_wf_order.
  Inversion left_wf.
  Inversion right_wf0; Assumption.

  Assumption.

  Inversion left_bounded.
  Inversion right_bounded0.
  Assumption.

  Assumption.

  Unfold create.
  Constructor.
  Apply wf_order_overbound with k:=k0.
  Inversion left_wf.
  Inversion right_bounded0.
  Assumption.

  Inversion left_wf.
  Assumption.

  Inversion left_wf.
  Inversion right_bounded0.
  Apply compare_reverse1.
  Assumption.

  Inversion left_wf.
  Inversion right_wf0.
  Assumption.

  Unfold create; Constructor.
  Inversion left_wf.
  Inversion right_wf0.
  Assumption.

  Inversion left_bounded.
  Inversion right_bounded0.
  Apply compare_reverse2; Assumption.

  Apply wf_order_underbound with k:=k.
  Inversion left_bounded.
  Inversion right_bounded0.
  Assumption.

  Assumption.
Qed.

Lemma bal_bounded_down :
  (k : key)
  (l : t)
  (k_add: key) (data_add : a)
  (r : t)
  (Horder : (compare k_add k) = GREATER)
  (Hl : (bounded_down k l))
  (Hr : (bounded_down k r))
  (bounded_down k (bal l k_add data_add r)).
Proof.
  Intros.
  Unfold bal.
  NewDestruct (le_gt_dec (height l) (S (S (height r)))).
  NewDestruct (le_gt_dec (height r) (S (S (height l)))).
  Apply create_bounded_down; Assumption.

  NewDestruct r.
  Absurd (gt O (S (S (height l)))).
  Auto with arith.

  Exact g.

  Simpl.
  NewDestruct (le_gt_dec (height left) (height right)).
  Apply create_bounded_down.
  Inversion Hr.
  Assumption.

  Apply create_bounded_down.
  Assumption.

  Assumption.

  Inversion Hr.
  Assumption.

  Inversion Hr.
  Assumption.

  NewDestruct left.
  Absurd (gt O (height right)).
  Auto with arith.

  Exact g0.

  Simpl.
  Apply create_bounded_down.
  Inversion Hr.
  Inversion left_bounded.
  Assumption.

  Apply create_bounded_down.
  Assumption.

  Assumption.

  Inversion Hr.
  Inversion left_bounded.
  Assumption.

  Apply create_bounded_down.
  Inversion Hr.
  Assumption.

  Inversion Hr.
  Inversion left_bounded.
  Assumption.

  Inversion Hr.
  Assumption.

  NewDestruct l.
  Absurd (gt O (S (S (height r)))).
  Auto with arith.

  Exact g.

  Simpl.
  NewDestruct (le_gt_dec (height right) (height left)).
  Apply create_bounded_down.
  Inversion Hl.
  Assumption.

  Inversion Hl.
  Assumption.

  Apply create_bounded_down.
  Assumption.

  Inversion Hl.
  Assumption.

  Assumption.

  NewDestruct right.
  Absurd (gt O (height left)).
  Auto with arith.

  Exact g0.

  Simpl.
  Apply create_bounded_down.
  Inversion Hl.
  Inversion right_bounded.
  Assumption.

  Apply create_bounded_down.
  Inversion Hl.
  Assumption.

  Inversion Hl.
  Assumption.

  Inversion Hl.
  Inversion right_bounded.
  Assumption.

  Apply create_bounded_down.
  Assumption.

  Inversion Hl.
  Inversion right_bounded.
  Assumption.

  Assumption.
Qed.

Lemma bal_bounded_up :
  (k : key)
  (l : t)
  (k_add: key) (data_add : a)
  (r : t)
  (Horder : (compare k_add k) = LESS)
  (Hl : (bounded_up k l))
  (Hr : (bounded_up k r))
  (bounded_up k (bal l k_add data_add r)).
Proof.
  Intros.
  Unfold bal.
  NewDestruct (le_gt_dec (height l) (S (S (height r)))).
  NewDestruct (le_gt_dec (height r) (S (S (height l)))).
  Apply create_bounded_up; Assumption.

  NewDestruct r.
  Absurd (gt O (S (S (height l)))).
  Auto with arith.

  Exact g.

  Simpl.
  NewDestruct (le_gt_dec (height left) (height right)).
  Apply create_bounded_up.
  Inversion Hr.
  Assumption.

  Apply create_bounded_up.
  Assumption.

  Assumption.

  Inversion Hr.
  Assumption.

  Inversion Hr.
  Assumption.

  NewDestruct left.
  Absurd (gt O (height right)).
  Auto with arith.

  Exact g0.

  Simpl.
  Apply create_bounded_up.
  Inversion Hr.
  Inversion left_bounded.
  Assumption.

  Apply create_bounded_up.
  Assumption.

  Assumption.

  Inversion Hr.
  Inversion left_bounded.
  Assumption.

  Apply create_bounded_up.
  Inversion Hr.
  Assumption.

  Inversion Hr.
  Inversion left_bounded.
  Assumption.

  Inversion Hr.
  Assumption.

  NewDestruct l.
  Absurd (gt O (S (S (height r)))).
  Auto with arith.

  Exact g.

  Simpl.
  NewDestruct (le_gt_dec (height right) (height left)).
  Apply create_bounded_up.
  Inversion Hl.
  Assumption.

  Inversion Hl.
  Assumption.

  Apply create_bounded_up.
  Assumption.

  Inversion Hl.
  Assumption.

  Assumption.

  NewDestruct right.
  Absurd (gt O (height left)).
  Auto with arith.

  Exact g0.

  Simpl.
  Apply create_bounded_up.
  Inversion Hl.
  Inversion right_bounded.
  Assumption.

  Apply create_bounded_up.
  Inversion Hl.
  Assumption.

  Inversion Hl.
  Assumption.

  Inversion Hl.
  Inversion right_bounded.
  Assumption.

  Inversion Hl.
  Inversion right_bounded.
  Apply create_bounded_up.
  Assumption.

  Assumption.

  Assumption.
Qed.

Lemma find_bal_exact :
  (left : t) (k : key) (data : a) (right : t)
  (left_wf : (wf_order left))
  (right_wf : (wf_order right))
  (left_bounded : (bounded_up k left))
  (right_bounded : (bounded_down k right))
  (kfind : key)
  (Hcompare : (compare k kfind) = EQUAL)
  ((coqfind (bal left k data right) kfind) = (Some a data)).
Proof.
  Intros.
  Unfold bal.
  NewDestruct (le_gt_dec (height left) (S (S (height right)))).
  NewDestruct (le_gt_dec (height right) (S (S (height left)))).
  Simpl.
  Replace (compare kfind k) with EQUAL.
  Trivial.

  Symmetry.
  Apply compare_equal_sym.
  Assumption.

  NewDestruct right.
  Absurd (gt O (S (S (height left)))).
  Auto with arith.

  Exact g.

  Simpl.
  NewDestruct (le_gt_dec (height left0) (height right)).
  Simpl.
  Replace (compare kfind k) with EQUAL.
  Inversion right_bounded.
  Replace (compare kfind k0) with LESS.
  Trivial.

  Symmetry.
  Apply compare_reverse1.
  Replace (compare k0 kfind) with (compare k0 k).
  Assumption.

  Apply compare_equiv2.
  Assumption.

  Symmetry.
  Apply compare_equal_sym.
  Assumption.

  NewDestruct left0.
  Simpl.
  Absurd (gt O (height right)).
  Auto with arith.

  Exact g0.

  Simpl.
  Replace (compare kfind k) with EQUAL.
  Replace (compare kfind k1) with (compare k k1).
  Replace (compare kfind k0) with (compare k k0).
  Inversion right_bounded.
  Inversion left_bounded0.
  Replace (compare k k1) with LESS.
  Trivial.

  Symmetry.
  Apply compare_reverse1.
  Trivial.

  Apply compare_equiv.
  Trivial.

  Apply compare_equiv.
  Trivial.

  Symmetry.
  Trivial.
  Apply compare_equal_sym.
  Assumption.

  NewDestruct left.
  Absurd (gt O (S (S (height right)))).
  Auto with arith.

  Exact g.

  Simpl.
  NewDestruct (le_gt_dec (height right0) (height left)).
  Simpl.
  Replace (compare kfind k0) with GREATER.
  Replace (compare kfind k) with EQUAL.
  Trivial.

  Symmetry.
  Apply compare_equal_sym.
  Assumption.

  Symmetry.
  Replace (compare kfind k0) with (compare k k0).
  Inversion left_bounded.
  Apply compare_reverse2.
  Assumption.

  Apply compare_equiv.
  Assumption.

  NewDestruct right0.
  Absurd (gt O (height left)).
  Auto with arith.

  Exact g0.

  Simpl.
  Replace (compare kfind k1) with (compare k k1).
  Replace (compare k k1) with GREATER.
  Replace (compare kfind k) with EQUAL.
  Trivial.

  Symmetry.
  Apply compare_equal_sym.
  Assumption.

  Symmetry.
  Inversion left_bounded.
  Inversion right_bounded0.
  Apply compare_reverse2.
  Assumption.

  Apply compare_equiv.
  Assumption.
Qed.

Lemma find_bal_less :
  (left : t) (k : key) (data : a) (right : t)
  (left_wf : (wf_order left))
  (right_wf : (wf_order right))
  (left_bounded : (bounded_up k left))
  (right_bounded : (bounded_down k right))
  (kfind : key)
  (Hcompare : (compare kfind k) = LESS)
  ((coqfind (bal left k data right) kfind) = (coqfind left kfind)).
Proof.
  Intros.
  Unfold bal.
  NewDestruct (le_gt_dec (height left) (S (S (height right)))).
  NewDestruct (le_gt_dec (height right) (S (S (height left)))).
  Simpl.
  Rewrite -> Hcompare.
  Trivial.

  NewDestruct right.
  Absurd (gt O (S (S (height left)))).
  Auto with arith.

  Exact g.

  Simpl.
  NewDestruct (le_gt_dec (height left0) (height right)).
  Simpl.
  Rewrite -> Hcompare.
  Inversion right_bounded.
  Replace (compare kfind k0) with LESS.
  Trivial.

  Symmetry.
  Apply compare_trans1 with y:=k.
  Assumption.

  Apply compare_reverse1.
  Assumption.

  NewDestruct left0.
  Absurd (gt O (height right)).
  Auto with arith.

  Exact g0.

  Simpl.
  Rewrite -> Hcompare.
  Replace (compare kfind k1) with LESS.
  Trivial.

  Symmetry.
  Inversion right_bounded.
  Apply compare_trans1 with y:=k.
  Assumption.

  Inversion left_bounded0.
  Apply compare_reverse1.
  Assumption.

  NewDestruct left.
  Absurd (gt O (S (S (height right)))).
  Auto with arith.

  Exact g.

  Simpl.
  NewDestruct (le_gt_dec (height right0) (height left)).
  Simpl.
  Rewrite -> Hcompare.
  Trivial.

  NewDestruct right0.
  Absurd (gt O (height left)).
  Auto with arith.

  Exact g0.

  Simpl.
  Inversion left_wf.
  Inversion left_bounded.
  Inversion right_bounded1.
  Inversion right_bounded0.
  Rewrite -> Hcompare.
  Cut (compare kfind k0)=EQUAL->(compare kfind k1)=EQUAL->False.
  Cut (compare kfind k0)=LESS->(compare kfind k1)=EQUAL->False.
  Cut (compare kfind k0)=LESS->(compare kfind k1)=GREATER->False.
  Cut (compare kfind k0)=EQUAL->(compare kfind k1)=GREATER->False.
  Intros HO1 HO2 HO3 HO4.
  NewDestruct (compare kfind k0).
  NewDestruct (compare kfind k1).
  Trivial.

  Apply False_ind.
  Apply HO3; Trivial.

  Apply False_ind.
  Apply HO2; Trivial.

  NewDestruct (compare kfind k1).
  Trivial.

  Apply False_ind.
  Apply HO4; Trivial.

  Apply False_ind.
  Apply HO1; Trivial.

  NewDestruct (compare kfind k1).
  Trivial.

  Trivial.

  Trivial.

  Intros HO1 HO2.
  Cut (compare k0 k1)=GREATER.
  Intro HO3.
  Apply compare_antisym2 with x:=k0 y:=k1.
  Assumption.

  Assumption.

  Replace (compare k0 k1) with (compare kfind k1).
  Assumption.

  Apply compare_equiv.
  Assumption.

  Intros HO1 HO2.
  Cut (compare kfind k1)=LESS.
  Rewrite -> HO2.
  Intro absurd.
  Discriminate absurd.

  Apply compare_trans1 with y:=k0.
  Assumption.

  Apply compare_reverse1.
  Assumption.

  Intros HO1 HO2.
  Cut (compare kfind k1)=LESS.
  Rewrite -> HO2.
  Intro absurd.
  Discriminate absurd.

  Apply compare_trans1 with y:=k0.
  Assumption.

  Apply compare_reverse1.
  Assumption.

  Intros HO1 HO2.
  Cut (compare k1 k0)=EQUAL.
  Rewrite -> key_bounded1.
  Intro absurd.
  Discriminate absurd.

  Replace (compare k1 k0) with (compare kfind k0).
  Assumption.

  Apply compare_equiv.
  Assumption.
Qed.

Lemma find_bal_greater :
  (left : t) (k : key) (data : a) (right : t)
  (left_wf : (wf_order left))
  (right_wf : (wf_order right))
  (left_bounded : (bounded_up k left))
  (right_bounded : (bounded_down k right))
  (kfind : key)
  (Hcompare : (compare kfind k) = GREATER)
  ((coqfind (bal left k data right) kfind) = (coqfind right kfind)).
Proof.
  Intros.
  Unfold bal.
  NewDestruct (le_gt_dec (height left) (S (S (height right)))).
  NewDestruct (le_gt_dec (height right) (S (S (height left)))).
  Simpl.
  Rewrite -> Hcompare.
  Trivial.

  NewDestruct right.
  Absurd (gt O (S (S (height left)))).
  Auto with arith.

  Exact g.

  Simpl.
  NewDestruct (le_gt_dec (height left0) (height right)).
  Simpl.
  Rewrite -> Hcompare.
  Trivial.

  NewDestruct left0.
  Absurd (gt O (height right)).
  Auto with arith.

  Exact g0.

  Simpl.
  Rewrite -> Hcompare.
  Inversion right_wf.
  Inversion left_bounded0.
  Inversion right_bounded.
  Inversion left_bounded2.
  Cut (compare kfind k1)=EQUAL->(compare kfind k0)=EQUAL->False.
  Cut (compare kfind k1)=EQUAL->(compare kfind k0)=GREATER->False.
  Cut (compare kfind k1)=LESS->(compare kfind k0)=EQUAL->False.
  Cut (compare kfind k1)=LESS->(compare kfind k0)=GREATER->False.
  Intros HO1 HO2 HO3 HO4.
  NewDestruct (compare kfind k1).
  NewDestruct (compare kfind k0).
  Trivial.

  Apply False_ind.
  Apply HO2; Trivial.

  Trivial.
  Apply False_ind.
  Apply False_ind.
  Apply HO1; Trivial.

  NewDestruct (compare kfind k0).
  Trivial.

  Apply False_ind.
  Apply HO4; Trivial.

  Apply False_ind.
  Apply HO3; Trivial.

  NewDestruct (compare kfind k0).
  Trivial.

  Trivial.

  Trivial.

  Intros HO1 HO2.
  Cut (compare k0 k1)=LESS.
  Intro HO3.
  Apply compare_antisym with x:=k0 y:=k1; Assumption.

  Apply compare_trans1 with y:=kfind.
  Apply compare_reverse1.
  Assumption.

  Assumption.

  Intros HO1 HO2.
  Cut (compare k0 k1)=LESS.
  Intro HO3.
  Apply compare_antisym with x:=k0 y:=k1; Assumption.

  Replace (compare k0 k1) with (compare kfind k1).
  Assumption.

  Apply compare_equiv.
  Assumption.

  Intros HO1 HO2.
  Cut (compare k0 k1)=LESS.
  Intro HO3.
  Apply compare_antisym with x:=k0 y:=k1; Assumption.

  Replace (compare k0 k1) with (compare k0 kfind).
  Apply compare_reverse1.
  Assumption.

  Apply compare_equiv2.
  Assumption.

  Intros HO1 HO2.
  Cut (compare k1 k0)=EQUAL.
  Rewrite -> key_bounded.
  Intro absurd.
  Discriminate absurd.

  Replace (compare k1 k0) with (compare kfind k0).
  Assumption.

  Apply compare_equiv.
  Assumption.

  NewDestruct left.
  Absurd (gt O (S (S (height right)))).
  Auto with arith.

  Exact g.

  Simpl.
  NewDestruct (le_gt_dec (height right0) (height left)).
  Simpl.
  Rewrite -> Hcompare.
  Inversion left_bounded.
  Replace (compare kfind k0) with GREATER.
  Trivial.

  Symmetry.
  Apply compare_reverse2.
  Apply compare_trans1 with y:=k.
  Assumption.

  Apply compare_reverse1.
  Assumption.

  NewDestruct right0.
  Absurd (gt O (height left)).
  Auto with arith.

  Exact g0.

  Simpl.
  Rewrite -> Hcompare.
  Inversion left_bounded.
  Inversion right_bounded0.
  Replace (compare kfind k1) with GREATER.
  Trivial.

  Symmetry.
  Apply compare_reverse2.
  Apply compare_trans1 with y:=k.
  Assumption.

  Apply compare_reverse1.
  Assumption.
Qed.

Lemma find_bal_imm :
  (left : t) (k : key) (data : a) (right : t)
  (left_wf : (wf_order left))
  (right_wf : (wf_order right))
  (left_bounded : (bounded_up k left))
  (right_bounded : (bounded_down k right))
  (kfind : key)
  (Hcompare : (compare kfind k) = EQUAL)
  ((coqfind (bal left k data right) kfind) = (Some a data)).
Proof.
  Intros.
  Unfold bal.
  NewDestruct (le_gt_dec (height left) (S (S (height right)))).
  NewDestruct (le_gt_dec (height right) (S (S (height left)))).
  Simpl.
  Rewrite -> Hcompare.
  Trivial.

  NewDestruct right.
  Absurd (gt O (S (S (height left)))).
  Auto with arith.

  Exact g.

  Simpl.
  NewDestruct (le_gt_dec (height left0) (height right)).
  Unfold create.
  Simpl.
  Inversion right_bounded.
  Rewrite -> Hcompare.
  Replace (compare kfind k0) with LESS.
  Trivial.

  Symmetry.
  Apply compare_reverse1.
  Replace (compare k0 kfind) with (compare k0 k).
  Assumption.

  Apply compare_equiv2.
  Apply compare_equal_sym.
  Assumption.

  NewDestruct left0.
  Absurd (gt O (height right)).
  Auto with arith.

  Exact g0.

  Simpl.
  Rewrite -> Hcompare.
  Inversion right_bounded.
  Inversion left_bounded0.
  Replace (compare kfind k1) with LESS.
  Trivial.

  Symmetry.
  Apply compare_reverse1.
  Replace (compare k1 kfind) with (compare k1 k).
  Assumption.

  Apply compare_equiv2.
  Apply compare_equal_sym.
  Assumption.

  NewDestruct left.
  Absurd (gt O (S (S (height right)))).
  Auto with arith.

  Exact g.

  Simpl.
  NewDestruct (le_gt_dec (height right0) (height left)).
  Simpl.
  Rewrite -> Hcompare.
  Inversion left_bounded.
  Replace (compare kfind k0) with GREATER.
  Trivial.

  Symmetry.
  Apply compare_reverse2.
  Replace (compare k0 kfind) with (compare k0 k).
  Assumption.

  Apply compare_equiv2.
  Apply compare_equal_sym.
  Assumption.

  NewDestruct right0.
  Absurd (gt O (height left)).
  Auto with arith.

  Exact g0.

  Simpl.
  Rewrite -> Hcompare.
  Inversion left_bounded.
  Inversion right_bounded0.
  Replace (compare kfind k1) with GREATER.
  Trivial.

  Symmetry.
  Apply compare_reverse2.
  Replace (compare k1 kfind) with (compare k1 k).
  Assumption.

  Apply compare_equiv2.
  Apply compare_equal_sym.
  Assumption.
Qed.
