(*
Functional maps as balanced binary trees

D. Monniaux, 2003
Laboratoire d'Informatique de l'Ecole Normale Superieure

$Id: map_map.v,v 1.9 2003/11/12 20:49:06 monniaux Exp $
*)

Section map_f.
  Variable a: Set.
  Variable b: Set.
  Variable f: a->b.

  Fixpoint map [tree : (t a)] : (t b) :=
    Cases tree of
    | Empty => (Empty b)
    | (Node l x d r h) => (Node b (map l) x (f d) (map r) h)
    end.

  Theorem map_correct : (tree : (t a)) (k_find : key)
    (coqfind b (map tree) k_find) =
    Cases (coqfind a tree k_find) of
    | None => (None b)
    | (Some x) => (Some b (f x))
    end.
  Proof.
    Intro tree.
    Elim tree.
    Simpl.
    Auto.

    Intros l Hl x d r Hr h k_find.
    Simpl.
    NewDestruct (compare k_find x).
    Apply Hl.

    Trivial.

    Apply Hr.
  Qed.

  Lemma map_bounded_up : (tree : (t a)) (y : key)
    (bounded : (bounded_up a y tree))
    (bounded_up b y (map tree)).
  Proof.
    Intro.
    Elim tree.
    Simpl.
    Intros.
    Constructor.

    Intros.
    Simpl.
    Inversion bounded.
    Constructor.
    Apply H; Assumption.

    Assumption.

    Apply H0; Assumption.
  Qed.

  Lemma map_bounded_down : (tree : (t a)) (y : key)
    (bounded : (bounded_down a y tree))
    (bounded_down b y (map tree)).
  Proof.
    Intro.
    Elim tree.
    Simpl.
    Intros.
    Constructor.

    Intros.
    Simpl.
    Inversion bounded.
    Constructor.
    Apply H; Assumption.

    Assumption.

    Apply H0; Assumption.
  Qed.

  Lemma map_wf_order : (tree : (t a))
    (tree_wf_order : (wf_order a tree))
    (wf_order b (map tree)).
  Proof.
    Intro.
    Elim tree.
    Simpl.
    Intros.
    Constructor.

    Intros.
    Simpl.
    Inversion tree_wf_order.
    Constructor.
    Apply H; Assumption.

    Apply H0; Assumption.

    Apply map_bounded_up; Assumption.

    Apply map_bounded_down; Assumption.
  Qed.

End map_f.

Section mapi_f.
  Variable a: Set.
  Variable b: Set.
  Variable f: key->a->b.

  Fixpoint mapi [tree : (t a)] : (t b) :=
    Cases tree of
    | Empty => (Empty b)
    | (Node l x d r h) => (Node b (mapi l) x (f x d) (mapi r) h)
    end.

  Variable f_respects : (k1, k2: key) (compare k1 k2)=EQUAL
    -> (z : a) (f k1 z)=(f k2 z).

  Theorem mapi_correct : (tree : (t a)) (k_find : key)
    (coqfind b (mapi tree) k_find) =
    Cases (coqfind a tree k_find) of
    | None => (None b)
    | (Some x) => (Some b (f k_find x))
    end.
  Proof.
    Intro tree.
    Elim tree.
    Simpl.
    Auto.

    Intros l Hl x d r Hr h k_find.
    Simpl.
    Cut
     (cmp:compare_result)
       (compare k_find x)=cmp
       ->Cases (compare k_find x) of
           LESS => (coqfind b (mapi l) k_find)
         | EQUAL => (Some b (f x d))
         | GREATER => (coqfind b (mapi r) k_find)
         end
          =Cases
             Cases (compare k_find x) of
               LESS => (coqfind a l k_find)
             | EQUAL => (Some a d)
             | GREATER => (coqfind a r k_find)
             end
           of
             (Some x0) => (Some b (f k_find x0))
           | None => (None b)
           end.
    Intro W; Apply W with cmp:=(compare k_find x); Trivial.

    Intros cmp Horder.
    Rewrite -> Horder.
    NewDestruct cmp.
    Apply Hl.

    Replace (f k_find d) with (f x d).
    Trivial.

    Apply f_respects.
    Apply compare_equal_sym; Assumption.

    Apply Hr.
  Qed.

  Lemma mapi_bounded_down : (tree : (t a)) (y : key)
    (bounded : (bounded_down a y tree))
    (bounded_down b y (mapi tree)).
  Proof.
    Intro.
    Elim tree.
    Simpl.
    Intros.
    Constructor.

    Intros.
    Simpl.
    Inversion bounded.
    Constructor.
    Apply H; Assumption.

    Assumption.

    Apply H0; Assumption.
  Qed.

  Lemma mapi_bounded_up : (tree : (t a)) (y : key)
    (bounded : (bounded_up a y tree))
    (bounded_up b y (mapi tree)).
  Proof.
    Intro.
    Elim tree.
    Simpl.
    Intros.
    Constructor.

    Intros.
    Simpl.
    Inversion bounded.
    Constructor.
    Apply H; Assumption.

    Assumption.

    Apply H0; Assumption.
  Qed.

  Lemma mapi_wf_order : (tree : (t a))
    (tree_wf_order : (wf_order a tree))
    (wf_order b (mapi tree)).
  Proof.
    Intro.
    Elim tree.
    Simpl.
    Intros.
    Constructor.

    Intros.
    Simpl.
    Inversion tree_wf_order.
    Constructor.
    Apply H; Assumption.

    Apply H0; Assumption.

    Apply mapi_bounded_up; Assumption.

    Apply mapi_bounded_down; Assumption.
  Qed.
End mapi_f.
