(*
Functional maps as balanced binary trees

D. Monniaux, 2003
Laboratoire d'Informatique de l'Ecole Normale Superieure
*)

Fixpoint min_binding [l : t] : key -> a -> (prod key a) := [x : key] [d : a]
  Cases l of
  | Empty => (pair key a x d)
  | (Node l2 x2 d2 _ _) => (min_binding l2 x2 d2)
  end.

Fixpoint remove_min_binding [l : t] : key -> a -> t -> t :=
  [x : key] [d : a] [r : t]
  Cases l of
  | Empty => r
  | (Node l2 x2 d2 r2 _) => (bal (remove_min_binding l2 x2 d2 r2) x d r)
  end.

Lemma remove_min_binding_bounded_up :
  (l : t) (x : key) (d : a) (r : t)
  (y : key)
  (x_bounded_up : (compare x y) = LESS)
  (yl_bounded_up : (bounded_up y l))
  (yr_bounded_up : (bounded_up y r))
  (bounded_up y (remove_min_binding l x d r)).
Proof.
  Intro.
  Elim l.
  Clear l.
  Simpl.
  Auto.

  Clear l.
  Intros.
  Simpl.
  Apply bal_bounded_up.
  Assumption.

  Apply H.
  Inversion yl_bounded_up.
  Assumption.

  Inversion yl_bounded_up; Assumption.

  Inversion yl_bounded_up; Assumption.

  Assumption.
Qed.
  
Lemma remove_min_binding_bounded_down :
  (l : t) (x : key) (d : a) (r : t)
  (y : key)
  (x_bounded_down : (compare x y) = GREATER)
  (yl_bounded_down : (bounded_down y l))
  (yr_bounded_down : (bounded_down y r))
  (bounded_down y (remove_min_binding l x d r)).
Proof.
  Intro.
  Elim l.
  Auto.

  Clear l.
  Intros.
  Simpl.
  Apply bal_bounded_down.
  Assumption.

  Apply H.
  Inversion yl_bounded_down; Assumption.

  Inversion yl_bounded_down; Assumption.

  Inversion yl_bounded_down; Assumption.

  Assumption.
Qed.

Lemma remove_min_binding_wf_order : (l : t) (x : key) (d : a) (r : t)
  (l_wf_order : (wf_order l)) (r_wf_order : (wf_order r))
  (l_bounded_up : (bounded_up x l))
  (r_bounded_down : (bounded_down x r))
  (wf_order (remove_min_binding l x d r)).
Proof.
  Intro.
  Elim l.
  Simpl.
  Auto.

  Intros.
  Simpl.
  Apply bal_wf_order.
  Apply H.
  Inversion l_wf_order; Assumption.

  Inversion l_wf_order; Assumption.

  Inversion l_wf_order; Assumption.

  Inversion l_wf_order; Assumption.

  Assumption.

  Apply remove_min_binding_bounded_up.
  Inversion l_bounded_up; Assumption.

  Inversion l_bounded_up; Assumption.

  Inversion l_bounded_up; Assumption.

  Assumption.
Qed.

Lemma coqfind_remove_min_binding : (l : t) (x : key) (d : a) (r : t)
  (l_wf_order : (wf_order l)) (r_wf_order : (wf_order r))
  (l_bounded_up : (bounded_up x l))
  (r_bounded_down : (bounded_down x r))
  (k_find : key)
  (Horder : (compare k_find (fst key a (min_binding l x d))) = GREATER)
  (coqfind (remove_min_binding l x d r) k_find) =
  (coqfind (create l x d r) k_find).
Proof.
  Intro.
  Elim l.
  Clear l.
  Simpl.
  Intros.
  Rewrite -> Horder.
  Trivial.

  Simpl.
  Intros nl Hnl nk nd nr Hnr h x d r.
  Intros wf_order_n wf_order_r bounded_up_x bounded_down_x k_find.
  Intro Horder.
  Cut
   (cmp1:compare_result)
     (compare k_find x)=cmp1
     ->(coqfind (bal (remove_min_binding nl nk nd nr) x d r) k_find)
        =Cases (compare k_find x) of
           LESS =>
            Cases (compare k_find nk) of
              LESS => (coqfind nl k_find)
            | EQUAL => (Some a nd)
            | GREATER => (coqfind nr k_find)
            end
         | EQUAL => (Some a d)
         | GREATER => (coqfind r k_find)
         end.
  Intro W; Apply W with cmp1:=(compare k_find x); Trivial.

  Intros cmp1 HO1.
  Rewrite -> HO1.
  NewDestruct cmp1.
  Rewrite -> find_bal_less.
  Apply Hnl.
  Inversion wf_order_n; Assumption.

  Inversion wf_order_n; Assumption.

  Inversion wf_order_n; Assumption.

  Inversion wf_order_n; Assumption.

  Assumption.

  Apply remove_min_binding_wf_order.
  Inversion wf_order_n; Assumption.

  Inversion wf_order_n; Assumption.

  Inversion wf_order_n; Assumption.

  Inversion wf_order_n; Assumption.

  Inversion wf_order_n; Assumption.

  Apply remove_min_binding_bounded_up.
  Inversion bounded_up_x; Assumption.

  Inversion bounded_up_x; Assumption.

  Inversion bounded_up_x; Assumption.

  Inversion bounded_up_x; Assumption.

  Assumption.

  Apply find_bal_imm.
  Apply remove_min_binding_wf_order.
  Inversion wf_order_n; Assumption.

  Inversion wf_order_n; Assumption.

  Inversion wf_order_n; Assumption.

  Inversion wf_order_n; Assumption.

  Inversion wf_order_n; Assumption.

  Apply remove_min_binding_bounded_up.
  Inversion bounded_up_x; Assumption.

  Inversion bounded_up_x; Assumption.

  Inversion bounded_up_x; Assumption.

  Inversion bounded_up_x; Assumption.

  Assumption.

  Rewrite -> find_bal_greater.
  Trivial.

  Apply remove_min_binding_wf_order.
  Inversion wf_order_n; Assumption.

  Inversion wf_order_n; Assumption.

  Inversion wf_order_n; Assumption.

  Inversion wf_order_n; Assumption.

  Inversion wf_order_n; Assumption.

  Apply remove_min_binding_bounded_up.
  Inversion bounded_up_x; Assumption.

  Inversion bounded_up_x; Assumption.

  Inversion bounded_up_x; Assumption.

  Inversion bounded_up_x; Assumption.

  Assumption.
Qed.

Lemma min_binding_bounded_up :
  (l : t) (x : key) (d : a)
  (l_wf_order : (wf_order l))
  (l_bounded_up : (bounded_up x l))
  (y : key)
  (yx_bounded_up : (compare y x) = GREATER)
  (compare y (fst key a (min_binding l x d))) = GREATER.
Proof.
  Intro.
  Elim l.
  Simpl.
  Auto.

  Clear l.
  Intros.
  Simpl.
  Apply H.
  Inversion l_wf_order; Assumption.

  Inversion l_wf_order; Assumption.

  Inversion l_bounded_up.
  Apply compare_reverse2.
  Apply compare_trans1 with y:=x.
  Assumption.

  Apply compare_reverse1.
  Assumption.
Qed.

Lemma min_binding_bounded_down :
  (l : t) (x : key) (d : a)
  (l_wf_order : (wf_order l))
  (l_bounded_up : (bounded_up x l))
  (y : key)
  (yl_bounded_down : (bounded_down y l))
  (yx_bounded_down : (compare y x) = LESS)
  (compare y (fst key a (min_binding l x d))) = LESS.
Proof.
  Intro l.
  Elim l.
  Simpl.
  Auto.

  Clear l.
  Intros.
  Simpl.
  Apply H.
  Inversion l_wf_order; Assumption.

  Inversion l_wf_order; Assumption.

  Inversion yl_bounded_down; Assumption.

  Inversion yl_bounded_down.
  Apply compare_reverse1.
  Assumption.
Qed.

Lemma min_binding_less_or_equal_than_node :
  (l : t) (x : key) (d : a)
  (l_wf_order : (wf_order l))
  (l_bounded_up : (bounded_up x l))
   (compare (fst key a (min_binding l x d)) x)=LESS \/
   (compare (fst key a (min_binding l x d)) x)=EQUAL.
Proof.
  Intro l.
  Elim l.
  Simpl.
  Intros.
  Right.
  Apply compare_refl.

  Intros.
  Left.
  Simpl.
  Apply compare_reverse1.
  Apply min_binding_bounded_up.
  Inversion l_wf_order; Assumption.

  Inversion l_wf_order; Assumption.

  Inversion l_bounded_up.
  Apply compare_reverse2.
  Assumption.
Qed.

Lemma bounded_down_from_min_binding :
  (l : t) (x : key) (d : a)
  (l_wf_order : (wf_order l))
  (y : key)
  (Horder : (compare y (fst key a (min_binding l x d))) = LESS)
  (bounded_down y l).
Proof.
  Intro l.
  Elim l.
  Simpl.
  Intros; Constructor.

  Simpl.
  Intros.
  Constructor.
  Apply H with x:=k d:=data.
  Inversion l_wf_order; Assumption.

  Inversion l_wf_order; Assumption.

  Apply compare_reverse2.
  Cut
   (compare (fst key a (min_binding left k data)) k)=LESS\/(compare
                                                             (fst key a
                                                               (min_binding
                                                                left k
                                                                data)) k)
                                                            =EQUAL.
  Intro choice.
  Case choice.
  Clear choice.
  Intro Horder2.
  Apply compare_trans1 with y:=(fst key a (min_binding left k data));
   Assumption.

  Clear choice.
  Intro Horder2.
  Replace (compare y k) with
  (compare y (fst key a (min_binding left k data))).
  Assumption.

  Apply compare_equiv2.
  Assumption.

  Apply min_binding_less_or_equal_than_node.
  Inversion l_wf_order; Assumption.

  Inversion l_wf_order; Assumption.

  Cut
   (compare (fst key a (min_binding left k data)) k)=LESS\/(compare
                                                             (fst key a
                                                               (min_binding
                                                                left k
                                                                data)) k)
                                                            =EQUAL.
  Intro choice.
  Case choice.
  Clear choice.
  Intro less.
  Apply wf_order_underbound with k:=k.
  Apply compare_trans1 with y:=(fst key a (min_binding left k data)).
  Assumption.

  Assumption.

  Inversion l_wf_order; Assumption.

  Clear choice.
  Intro equal.
  Apply wf_order_underbound with k:=k.
  Replace (compare y k) with
  (compare y (fst key a (min_binding left k data))).
  Assumption.

  Apply compare_equiv2.
  Assumption.

  Inversion l_wf_order; Assumption.

  Apply min_binding_less_or_equal_than_node.
  Inversion l_wf_order; Assumption.

  Inversion l_wf_order; Assumption.
Qed.


Lemma min_binding_bounds_down_removed :
  (l : t) (x : key) (d : a) (r : t)
  (l_bounded_up : (bounded_up x l))
  (r_bounded_down : (bounded_down x r))
  (l_wf_order : (wf_order l))
  (r_wf_order : (wf_order l))
  (bounded_down (fst key a (min_binding l x d)) (remove_min_binding l x d r)).
Proof.
  Intro l.
  Elim l.
  Simpl.
  Auto.

  Clear l.
  Intros.
  Simpl.
  Apply bal_bounded_down.
  Apply min_binding_bounded_up.
  Inversion l_wf_order; Assumption.

  Inversion l_wf_order; Assumption.

  Inversion l_bounded_up.
  Apply compare_reverse2.
  Assumption.

  Apply H.
  Inversion l_wf_order; Assumption.

  Inversion l_wf_order; Assumption.

  Inversion l_wf_order; Assumption.

  Inversion l_wf_order; Assumption.

  Apply wf_order_underbound with k:=x.
  Apply compare_reverse1.
  Apply min_binding_bounded_up.
  Inversion l_wf_order; Assumption.

  Inversion l_wf_order; Assumption.

  Inversion l_bounded_up.
  Apply compare_reverse2.
  Assumption.

  Assumption.
Qed.

Lemma min_binding_finds_binding2 :
  (l : t) (x : key) (d : a)
  (l_wf_order : (wf_order l))
  (l_bounded_up : (bounded_up x l))
   (Horder: (compare (fst key a (min_binding l x d)) x)=EQUAL)
    (snd key a (min_binding l x d)) =d.
Proof.
  Intro l.
  NewDestruct l.
  Simpl.
  Auto.

  Simpl.
  Intros.
  Cut (compare x (fst key a (min_binding left k data)))=GREATER.
  Intro Z.
  Cut (compare (fst key a (min_binding left k data)) x)=LESS.
  Rewrite -> Horder.
  Intro absurd; Discriminate absurd.

  Apply compare_reverse1.
  Assumption.

  Apply min_binding_bounded_up.
  Inversion l_wf_order; Assumption.

  Inversion l_wf_order; Assumption.

  Inversion l_bounded_up.
  Apply compare_reverse2.
  Assumption.
Qed.

Lemma min_binding_finds_binding1 :
  (l : t) (x : key) (d : a)
  (l_wf_order : (wf_order l))
  (l_bounded_up : (bounded_up x l))
   (Horder: (compare (fst key a (min_binding l x d)) x)=LESS)
    (coqfind l (fst key a (min_binding l x d))) =
    (Some a (snd key a (min_binding l x d))).
Proof.
  Intro l.
  Elim l.
  Clear l.
  Simpl.
  Intros.
  Cut (compare x x)=EQUAL.
  Rewrite -> Horder.
  Intro Z; Discriminate Z.

  Apply compare_refl.

  Intros tl Htl tk td tr Htr h x d t_wf_order t_bounded_up Horder.
  Simpl.
  Cut
   (cmp1:compare_result)
     (compare (fst key a (min_binding tl tk td)) tk)=cmp1
     ->Cases (compare (fst key a (min_binding tl tk td)) tk) of
         LESS => (coqfind tl (fst key a (min_binding tl tk td)))
       | EQUAL => (Some a td)
       | GREATER => (coqfind tr (fst key a (min_binding tl tk td)))
       end=(Some a (snd key a (min_binding tl tk td))).
  Intro W.
  Apply W with cmp1:=(compare (fst key a (min_binding tl tk td)) tk);
   Trivial.

  Intros cmp1 HO1.
  NewDestruct cmp1.
  Rewrite -> HO1.
  Apply Htl.
  Inversion t_wf_order; Assumption.

  Inversion t_wf_order; Assumption.

  Assumption.

  Rewrite -> HO1.
  Replace (snd key a (min_binding tl tk td)) with td; Trivial.
  Symmetry.
  Apply min_binding_finds_binding2.
  Inversion t_wf_order; Assumption.

  Inversion t_wf_order; Assumption.

  Assumption.

  Rewrite -> HO1.
  Cut
   (compare (fst key a (min_binding tl tk td)) tk)=LESS\/(compare
                                                           (fst key a
                                                             (min_binding
                                                               tl tk td))
                                                           tk)=EQUAL.
  Intro split.
  Case split; Rewrite -> HO1; Intro Z; Discriminate Z.

  Apply min_binding_less_or_equal_than_node.
  Inversion t_wf_order; Assumption.

  Inversion t_wf_order; Assumption.
Qed.
