(*
Functional maps as balanced binary trees

D. Monniaux, 2003
Laboratoire d'Informatique de l'Ecole Normale Superieure
*)

Inductive bounded_up [high : key] : t->Prop :=
| empty_bounded_up : (bounded_up high Empty)
| node_bounded_up :
   (left : t) (k : key) (data : a) (right : t) (height : nat)
   (left_bounded : (bounded_up high left))
   (key_bounded : ((compare k high) = LESS))
   (right_bounded : (bounded_up high right))
   (bounded_up high (Node left k data right height)).

Inductive bounded_down [low : key] : t->Prop :=
| empty_bounded_down : (bounded_down low Empty)
| node_bounded_down :
   (left : t) (k : key) (data : a) (right : t) (height : nat)
   (left_bounded : (bounded_down low left))
   (key_bounded : ((compare k low) = GREATER))
   (right_bounded : (bounded_down low right))
   (bounded_down low (Node left k data right height)).

Inductive wf_order : t->Prop :=
| empty_wf_order : (wf_order Empty)
| node_wf_order :
   (left : t) (k : key) (data : a) (right : t) (height : nat)
   (left_wf : (wf_order left))
   (right_wf : (wf_order right))
   (left_bounded : (bounded_up k left))
   (right_bounded : (bounded_down k right))
   (wf_order (Node left k data right height)).

Lemma wf_order_overbound : (k, ko: key) ((compare ko k) = GREATER) -> 
  (v : t) (bounded_up k v) -> (bounded_up ko v).
Proof.
  Intros k ko rel v.
  NewInduction v.
  Intro _.
  Apply empty_bounded_up.

  Intro Z.
  Inversion Z.
  Apply node_bounded_up.
  Apply IHv1.
  Assumption.

  Apply compare_trans1 with x:=k0 y:=k z:=ko.
  Assumption.

  Replace LESS with (invert_compare_result (compare ko k)).
  Rewrite <- compare_invert.
  Trivial.

  Rewrite -> rel.
  Simpl.
  Trivial.

  Apply IHv2.
  Assumption.
Qed.

Lemma wf_order_underbound : (k, ko: key) ((compare ko k) = LESS) -> 
  (v : t) (bounded_down k v) -> (bounded_down ko v).
Proof.
  Intros k ko rel v.
  NewInduction v.
  Intro _.
  Constructor.

  Intro Z.
  Inversion Z.
  Constructor.
  Apply IHv1.
  Assumption.

  Apply compare_reverse2.
  Apply compare_trans1 with y:=k.
  Assumption.

  Apply compare_reverse1.
  Assumption.

  Apply IHv2.
  Assumption.
Qed.

Lemma wf_order_overbound_equal :
  (k, ko: key) ((compare ko k) = EQUAL) -> 
  (v : t) (bounded_up k v) -> (bounded_up ko v).
Proof.
  Intros k ko rel v.
  NewInduction v.
  Intro _.
  Apply empty_bounded_up.

  Intro Z.
  Inversion Z.
  Apply node_bounded_up.
  Apply IHv1.
  Assumption.

  Replace (compare k0 ko) with (compare k0 k).
  Assumption.

  Apply compare_equiv2.
  Apply compare_equal_sym.
  Assumption.

  Apply IHv2.
  Assumption.
Qed.

Lemma wf_order_underbound_equal :
  (k, ko: key) ((compare ko k) = EQUAL) -> 
  (v : t) (bounded_down k v) -> (bounded_down ko v).
Proof.
  Intros k ko rel v.
  NewInduction v.
  Intro _.
  Apply empty_bounded_down.

  Intro Z.
  Inversion Z.
  Apply node_bounded_down.
  Apply IHv1.
  Assumption.

  Replace (compare k0 ko) with (compare k0 k).
  Assumption.

  Apply compare_equiv2.
  Apply compare_equal_sym.
  Assumption.

  Apply IHv2.
  Assumption.
Qed.
